#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Copyright (C) 2020 Christoph Fink, University of Helsinki
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 3
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, see <http://www.gnu.org/licenses/>.


"""
Download (all) georeferenced flickr posts.

Overloaded to provide fancy console output
"""


__all__ = ["FancyFlickrHistoryDownloader"]


import threading

import blessed

from .basicflickrhistorydownloader import BasicFlickrHistoryDownloader
from . import __version__ as version


class FancyFlickrHistoryDownloader(BasicFlickrHistoryDownloader):
    """
    Download (all) georeferenced flickr posts.

    With fancy console output.
    """

    WELCOME = (
        "{t.bold}{t.blue} ### flickrhistory "
        + "{t.normal}{t.blue}"
        + version
        + "{t.bold} ###"

        + "{t.normal}"
    )

    STATUS = (
        "{t.normal} Downloaded metadata for "
        + "{t.bold}{t.magenta}{photos: 9d} 📷 photos "
        + "{t.normal}{t.magenta}{photo_rate: 11.1f}/s\n"

        + "{t.normal} and updated             "
        + "{t.bold}{t.red}{profiles: 9d} 👱 user profiles "
        + "{t.normal}{t.red}{profile_rate: 3.1f}/s\n"

        + "{t.normal} using                   "
        + "{t.bold}{t.green}{workers: 9d} 💪 workers\n"

        + "{t.normal}{t.bold} TODO:                {todo: 12d} 🚧 time slots"

        + "{t.normal}"
    )
    STATUS_LINES = len(STATUS.splitlines())

    SHUTDOWN_ANNOUNCEMENT = "{t.bold}Cleaning up. 🛑 {t.normal}"

    SUMMARY = (
        "{t.normal}Downloaded  {t.bold}{t.magenta}{photos: 9d} 📷 photos "
        + "{t.normal}{t.magenta}{photo_rate: 11.1f}/s\n"
        + "{t.normal}and updated {t.bold}{t.red}{profiles: 9d} 👱 user profiles "
        + "{t.normal}{t.red}{profile_rate: 3.1f}/s\n"

        + "{t.normal}"
    )

    def __init__(self):
        """Initialise FlickrHistoryDownloader."""
        super().__init__()

        self.terminal = blessed.Terminal()

        print(self.WELCOME.format(t=self.terminal))

        # scroll down terminal, in case we’re at the bottom
        print(self.STATUS_LINES * "\n", end="")

        self.pos_y, _ = self.terminal.get_location(timeout=5)
        self._photo_count = 0
        self._profile_count = 0

    def report_progress(self):
        """Report current progress."""
        photo_count, photo_rate, profile_count, profile_rate = self._statistics

        with self.terminal.location(0, (self.pos_y - self.STATUS_LINES)):
            print(
                self.STATUS.format(
                    t=self.terminal,
                    photos=photo_count,
                    photo_rate=photo_rate,
                    profiles=profile_count,
                    profile_rate=profile_rate,
                    workers=(threading.active_count() - self.NUM_MANAGERS),
                    todo=len(self._todo_deque)
                )
            )

    def announce_shutdown(self):
        """Tell the user that we initiated shutdown."""
        # clear the status output
        for i in range(self.STATUS_LINES):
            with self.terminal.location(0, (self.pos_y - (i + 1))):
                print(self.terminal.clear_eol)

        with self.terminal.location(0, (self.pos_y - self.STATUS_LINES)):
            print(self.SHUTDOWN_ANNOUNCEMENT.format(t=self.terminal))

    def summarise_overall_progress(self):
        """
        Summarise what we have done.

        (Called right before exit)
        """
        photo_count, photo_rate, profile_count, profile_rate = self._statistics
        with self.terminal.location(0, (self.pos_y - self.STATUS_LINES)):
            print(
                self.SUMMARY.format(
                    t=self.terminal,
                    photos=photo_count,
                    photo_rate=photo_rate,
                    profiles=profile_count,
                    profile_rate=profile_rate
                )
            )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (C) 2018 Christoph Fink, University of Helsinki
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 3
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, see <http://www.gnu.org/licenses/>.


"""An exception to handle receiving SIGTERM signals."""


__all__ = ["SigTermReceivedException"]


import signal


class SigTermReceivedException(Exception):
    """Raised when SIGTERM signal received."""


def _received_sig_term(*args):
    raise SigTermReceivedException


signal.signal(signal.SIGTERM, _received_sig_term)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Copyright (C) 2020 Christoph Fink, University of Helsinki
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 3
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, see <http://www.gnu.org/licenses/>.


"""Worker threads wrapping an APIDownloader."""


__all__ = ["CacheUpdaterThread"]


import queue
import threading

from .cache import Cache


class CacheUpdaterThread(threading.Thread):
    """Wraps an ApiDownloader to run in a separate thread."""

    def __init__(
            self,
            done_queue
    ):
        """
        Intialize a CacheUpdaterThread.

        Args:
            done_queue: queue.Queue with updated TimeSpans
        """
        super().__init__()
        self._done_queue = done_queue
        self.shutdown = threading.Event()
        self.status = "init"

    def run(self):
        """Get TimeSpans off done_queue and update cache."""
        while True:
            try:
                newly_downloaded = self._done_queue.get(timeout=0.1)
                with Cache() as cache:
                    cache["already downloaded"] += newly_downloaded
                    self.status = "added {}".format(newly_downloaded)
            except queue.Empty:
                if self.shutdown.is_set():
                    break



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Copyright (C) 2020 Christoph Fink, University of Helsinki
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 3
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, see <http://www.gnu.org/licenses/>.


"""A period in time starting at a datetime and ending at another datetime."""


__all__ = [
    "TimeSpan"
]


import datetime

import yaml


class TimeSpan(yaml.YAMLObject):
    """A period in time starting at a datetime and ending at another datetime."""

    def __init__(
            self,
            start,
            end
    ):
        """Initialise a new TimeSpan object."""
        self.start = start
        self.end = end

    def __str__(self):
        """Return a string representation of this TimeSpan."""
        return (
            "<{:s}"
            + "({:%Y-%m-%dT%H:%M:%S.000Z}-{:%Y-%m-%dT%H:%M:%S.000Z})>"
        ).format(
            self.__class__.__name__,
            self.start,
            self.end
        )

    def __repr__(self):
        """Represent this TimeSpan in readable form."""
        return self.__str__()

    @property
    def start(self):
        """Return the start time of this time span."""
        return self._start

    @start.setter
    def start(self, start):
        if not isinstance(start, datetime.datetime):
            raise TypeError("Expected datetime.datetime")
        self._start = start

    @property
    def end(self):
        """Return the end time of this time span."""
        return self._end

    @end.setter
    def end(self, end):
        if not isinstance(end, datetime.datetime):
            raise TypeError("Expected datetime.datetime")
        self._end = end

    @property
    def duration(self):
        """Compute the time between TimeSpan.start and TimeSpan.end."""
        return self._end - self.start

    def __lt__(self, other):
        """Test if this TimeSpan starts earlier than another one."""
        return self.start < other.start

    def __gt__(self, other):
        """Test if this TimeSpan ends later than another one."""
        return self.end > other.end

    def __eq__(self, other):
        """Test if this TimeSpan and another one are equal."""
        return (
            self.start == other.start
            and self.end == other.end
        )

    def __add__(self, other):
        """Add another TimeSpan to this one."""
        if not isinstance(other, list):
            other = [other]

        # “recursively” merge all time spans in list
        not_yet_merged = sorted(other + [self])
        merged = [not_yet_merged.pop(0)]

        while not_yet_merged:
            # pop last member of merged time spans
            # and first one of still to be merged time spans
            last = merged.pop()
            new = not_yet_merged.pop(0)

            # check for overlap of the two,
            # then add one or two items to the merged list
            if last.end >= new.start:
                merged.append(
                    TimeSpan(
                        min(last.start, new.start),
                        max(last.end, new.end)
                    )
                )
            else:
                merged += [last, new]

        return merged

    def __radd__(self, other):
        """Add this TimeSpan to another one."""
        if other == 0:  # default case for calling `sum()`
            return self
        return self.__add__(other)

    def __truediv__(self, other):
        """Divide (split!) this TimeSpan into [other: int] pieces."""
        if not isinstance(other, int):
            raise TypeError("TimeSpan can only be divided into [int] pieces.")

        if self.duration == datetime.timedelta(0):
            return [self]

        pieces = []
        duration = self.duration / other
        for i in range(other - 1):
            pieces.append(
                TimeSpan(
                    self.start + i * duration,
                    self.start + (i + 1) * duration
                )
            )
        pieces.append(
            TimeSpan(
                self.start + (other - 1) * duration,
                self.end
            )
        )

        return pieces

    yaml_tag = "!TimeSpan"
    yaml_loader = yaml.SafeLoader
    yaml_dumper = yaml.SafeDumper

    @classmethod
    def from_yaml(cls, loader, node):
        """Convert a YAML !TimeSpan block to a TimeSpan object."""
        data = loader.construct_mapping(node)
        return TimeSpan(data["start"], data["end"])

    @classmethod
    def to_yaml(cls, dumper, data):
        """Map a TimeSpan object to its YAML representation."""
        return dumper.represent_mapping(
            "!TimeSpan",
            (
                ("start", data.start),
                ("end", data.end)
            )
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Copyright (C) 2020 Christoph Fink, University of Helsinki
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 3
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, see <http://www.gnu.org/licenses/>.


"""Worker threads wrapping an APIDownloader."""


__all__ = ["PhotoDownloaderThread"]


import threading
import time

import sqlalchemy

from .config import Config
from .databaseobjects import FlickrPhoto
from .exceptions import ApiResponseError, DownloadBatchIsTooLargeError
from .photodownloader import PhotoDownloader


class PhotoDownloaderThread(threading.Thread):
    """Wraps an PhotoDownloader to run in a separate thread."""

    def __init__(
            self,
            api_key_manager,
            todo_deque,
            done_queue
    ):
        """
        Intialize an PhotoDownloaderThread.

        Args:
            api_key_manager: instance of an ApiKeyManager
            todo_deque: collections.deque that serves TimeSpans
                        that need to be downloaded
            done_queue: queue.Queue into which to put TimeSpans
                        that have been downloaded

        """
        super().__init__()

        self.count = 0

        self._api_key_manager = api_key_manager
        self._todo_deque = todo_deque
        self._done_queue = done_queue

        self.shutdown = threading.Event()

        with Config() as config:
            self._engine = sqlalchemy.create_engine(
                config["database_connection_string"]
            )

    def run(self):
        """Get TimeSpans off todo_deque and download photos."""
        while not self.shutdown.is_set():
            try:
                timespan = self._todo_deque.pop()
            except IndexError:
                break

            photo_downloader = PhotoDownloader(timespan, self._api_key_manager)

            try:
                for photo in photo_downloader.photos:
                    with sqlalchemy.orm.Session(
                        self._engine
                    ) as session:
                        try:
                            with session.begin():
                                flickr_photo = (
                                    FlickrPhoto.from_raw_api_data_flickrphotossearch(
                                        photo
                                    )
                                )
                                session.merge(flickr_photo)
                        except sqlalchemy.exc.IntegrityError:
                            # remedy race conditions
                            # TODO: find out how to avoid them
                            time.sleep(1.0)
                            with session.begin():
                                session.flush()
                                flickr_photo = (
                                    FlickrPhoto.from_raw_api_data_flickrphotossearch(
                                        photo
                                    )
                                )
                                session.merge(flickr_photo)

                    self.count += 1

                    if self.shutdown.is_set():
                        # let’s only report back on how much we
                        # in fact downloaded, not what our quota was
                        timespan.end = flickr_photo.date_posted
                        break

            except ApiResponseError:
                # API returned some bogus/none-JSON data
                # let’s add this timespan to the other end
                # of the todo deque and start over
                # TODO: implement logging and log the
                # data (which is in the exception’s message)
                self._todo_deque.appendleft(timespan)
                continue

            except DownloadBatchIsTooLargeError:
                # too many photos in this time span,
                # let’s split it in half and re-inject
                # it to the todo deque
                for half_timespan in timespan / 2:
                    self._todo_deque.append(half_timespan)

                # get a new timespan from the deque :)
                continue

            # … report to parent thread how much we worked
            self._done_queue.put(timespan)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Copyright (C) 2020 Christoph Fink, University of Helsinki
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 3
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, see <http://www.gnu.org/licenses/>.


"""Download (all) georeferenced flickr posts."""


__all__ = ["BasicFlickrHistoryDownloader"]


import collections
import datetime
import queue
import math
import multiprocessing
import sys
import threading
import time

from .apikeymanager import ApiKeyManager
from .cache import Cache
from .cacheupdaterthread import CacheUpdaterThread
from .config import Config
from .photodownloaderthread import PhotoDownloaderThread
from .sigtermreceivedexception import SigTermReceivedException
from .timespan import TimeSpan
from .userprofileupdaterthread import UserProfileUpdaterThread


class BasicFlickrHistoryDownloader:
    """Download (all) georeferenced flickr posts."""

    NUM_WORKERS = multiprocessing.cpu_count() + 1  # 1 == user_profile_updater
    NUM_MANAGERS = 2  # main thread + cache_updater

    # if output into pipe (e.g. logger, systemd), then
    # print status every 10 minutes, else every 1/5 sec
    # also normal linefeed instead of carriage return for piped output
    STATUS_UPDATE_SEC = 0.2 if sys.stderr.isatty() else 600
    STATUS_UPDATE_LINE_END = "\r" if sys.stderr.isatty() else "\n"

    def __init__(self):
        """Intialise a FlickrHistory object."""
        self.started = datetime.datetime.now()

        self._todo_deque = collections.deque()
        self._done_queue = queue.Queue()

        self._worker_threads = []
        self._cache_updater_thread = CacheUpdaterThread(self._done_queue)

        with Config() as config:
            self._api_key_manager = ApiKeyManager(config["flickr_api_keys"])

    def download(self):
        """Download all georeferenced flickr posts."""
        for gap in self.gaps_in_download_history:
            self._todo_deque.append(gap)

        try:
            # start downloaders
            for _ in range(self.NUM_WORKERS):
                worker = PhotoDownloaderThread(
                    self._api_key_manager,
                    self._todo_deque,
                    self._done_queue
                )
                worker.start()
                self._worker_threads.append(worker)

            # start user profile updaters
            for i in range(self.NUM_WORKERS):
                worker = UserProfileUpdaterThread(
                    self._api_key_manager,
                    (i + 1, self.NUM_WORKERS)
                )
                worker.start()
                self._worker_threads.append(worker)

            # start cache updater
            self._cache_updater_thread = CacheUpdaterThread(self._done_queue)
            self._cache_updater_thread.start()

            while threading.active_count() > self.NUM_MANAGERS:
                self.report_progress()
                time.sleep(self.STATUS_UPDATE_SEC)

        except (
            KeyboardInterrupt,
            SigTermReceivedException
        ):
            self.announce_shutdown()
            for worker in self._worker_threads:
                worker.shutdown.set()

        finally:
            self.summarise_overall_progress()
            for worker in self._worker_threads:
                worker.join()
            self._cache_updater_thread.shutdown.set()
            self._cache_updater_thread.join()

    def report_progress(self):
        """Report current progress."""
        photo_count, _, profile_count, _ = self._statistics
        print(
            (
                "Downloaded metadata for {photos: 6d} photos "
                + "and {profiles: 4d} user profiles "
                + "using {workers:d} workers, "
                + "{todo:d} time slots to cover"
            ).format(
                photos=photo_count,
                profiles=profile_count,
                workers=(threading.active_count() - self.NUM_MANAGERS),
                todo=len(self._todo_deque)
            ),
            file=sys.stderr,
            end=self.STATUS_UPDATE_LINE_END,
            flush=True
        )

    def announce_shutdown(self):
        """Tell the user that we initiated shutdown."""
        print(
            "Cleaning up" + (" " * 69),  # 80 - len("Cleaning up")
            file=sys.stderr,
            end=self.STATUS_UPDATE_LINE_END,
            flush=True
        )

    def summarise_overall_progress(self):
        """
        Summarise what we have done.

        (Called right before exit)
        """
        photo_count, _, profile_count, _ = self._statistics
        print(
            (
                "Downloaded {photos:d} photos "
                + "and {profiles:d} user profiles"
            ).format(
                photos=photo_count,
                profiles=profile_count
            ),
            file=sys.stderr
        )

    @property
    def gaps_in_download_history(self):
        """Find gaps in download history."""
        already_downloaded = self.already_downloaded_timespans
        one_day = datetime.timedelta(days=1)  # for comparison

        for i in range(len(already_downloaded) - 1):
            gap = TimeSpan(
                already_downloaded[i].end,
                already_downloaded[i + 1].start
            )
            if gap.duration > one_day:
                divider = math.ceil(gap.duration / one_day)
                for part_of_gap in gap / divider:
                    yield part_of_gap
            else:
                yield gap

    @property
    def already_downloaded_timespans(self):
        """Figure out for which time spans we already have data."""
        with Cache() as cache:
            try:
                timespans = cache["already downloaded"]
            except KeyError:
                timespans = []

        # delete existing 0-length time spans
        timespans = [
            timespan for timespan in timespans
            if timespan.duration > datetime.timedelta(0)
        ]

        # add 0-length time spans for
        # - “beginning of all time”
        #  - now()
        #
        # beginning of time cannot be simply epoch 0, because the flickr API
        # matches dates slightly fuzzily, i.e. we would get ALL (or many) photos
        # that have a corrupted or missing upload date (don’t ask me, how flickr
        # managed to mess up the upload date)
        # on top of that, some small timestamps seems to be simply 0 +- timezone offset
        # which invalidates pretty much the entire first day after epoch 0
        # this is why we use epoch 0 + 1 day
        zero = (
            datetime.datetime.fromtimestamp(0, tz=datetime.timezone.utc)
            + datetime.timedelta(days=1)
        )
        now = datetime.datetime.now(datetime.timezone.utc)
        timespans += [
            TimeSpan(zero, zero),
            TimeSpan(now, now)
        ]

        return sum(timespans)  # sum resolves overlaps

    @property
    def _statistics(self):
        runtime = float((datetime.datetime.now() - self.started).total_seconds())

        photo_count = sum([
            worker.count
            for worker in self._worker_threads
            if isinstance(worker, PhotoDownloaderThread)
        ])
        photo_rate = photo_count / runtime

        profile_count = sum([
            worker.count
            for worker in self._worker_threads
            if isinstance(worker, UserProfileUpdaterThread)
        ])
        profile_rate = profile_count / runtime

        return (photo_count, photo_rate, profile_count, profile_rate)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Copyright (C) 2019 Christoph Fink, University of Helsinki
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 3
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, see <http://www.gnu.org/licenses/>.

"""Manages globally cached variables."""


__all__ = ["Cache"]


import os
import os.path
import warnings

import yaml


class YamlNoAliasDumper(yaml.SafeDumper):
    """YAML Dumper that does not write out aliases and anchors."""

    def ignore_aliases(self, data):
        return True


class Cache:
    """
    Global object holding variables cached in-between runs.

    Tries to load cache file from YAML files in default locations
    (/var/cache/{module}.yml, ~/.cache/{module}.yml,
    %LOCALAPPDATA%/{module}.yml, ${XDG_CACHE_HOME}/{module}.yml).

    """

    def __init__(
            self,
            cache=None,
            cache_file_basename=None
    ):
        """Initialise a Cache object, load cache from file."""
        self._cache = {}

        if cache_file_basename is None:
            cache_file_basename = self.__module__.split(".")[0]

        self._cache_file = os.path.abspath(
            os.path.join(
                (
                    os.environ.get('LOCALAPPDATA')
                    or os.environ.get('XDG_CACHE_HOME')
                    or os.path.join(os.environ['HOME'], '.cache')
                ),
                "{:s}.yml".format(cache_file_basename)
            )
        )

        self._cache = self._load_cache()
        if cache is not None:
            self._cache.update(cache)

    def _load_cache(self):
        cache = {}

        try:
            cache.update(
                yaml.safe_load(
                    open(self._cache_file, "r", encoding="utf-8")
                )
            )
        except FileNotFoundError:
            pass

        if cache == {}:
            warnings.warn(
                "No cache found in file {}, starting empty".format(
                    self._cache_file
                )
            )

        return cache

    def _save_cache(self):
        try:
            yaml.dump(
                self._cache,
                open(self._cache_file, "w", encoding="utf-8"),
                Dumper=YamlNoAliasDumper
            )
        except PermissionError:
            warnings.warn(
                "Could not write cache to {}".format(
                    self._cache_file
                )
            )

    def __getitem__(self, pos):
        """Retrieve a cache entry."""
        return self._cache[pos]

    def __setitem__(self, pos, value):
        """Set the value of a cache entry."""
        self._cache[pos] = value
        self._save_cache()  # don’t rely on this!
        # if you update items inside a dict,
        # __setitem__ is not called
        #
        # rather, use the context manager:
        # ```
        # with Cache() as cache:
        #     cache["key1"]["key2"] = "value"

    def __delitem__(self, pos):
        """Delete a cache entry."""
        del self._cache[pos]
        self._save_cache()

    def __iter__(self):
        """Iterate over all entries in the cache."""
        return iter(self._cache)

    def __enter__(self):
        """Enter cache context."""
        self._load_cache()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit cache context."""
        self._save_cache()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Copyright (C) 2020 Christoph Fink, University of Helsinki
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 3
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, see <http://www.gnu.org/licenses/>.


"""Download all data covering a time span from the flickr API."""


__all__ = ["UserProfileDownloader"]


import json

import requests
import urllib3

from .exceptions import ApiResponseError


class UserProfileDownloader:
    """Download user profile data from the flickr API."""

    API_ENDPOINT_URL = "https://api.flickr.com/services/rest/"

    def __init__(self, api_key_manager):
        """Intialize an PhotoDownloader."""
        self._api_key_manager = api_key_manager

    def get_profile_for_id_and_farm(self, user_id, farm):
        """Retrieve profile data by user_id and farm identifier."""
        return self.get_profile_for_nsid("@N0".join([user_id, farm]))

    def get_profile_for_nsid(self, nsid):
        """Get profile data by nsid."""
        query = {
            "method": "flickr.profile.getProfile",
            "format": "json",
            "nojsoncallback": 1,
            "user_id": nsid
        }

        params = {}
        with self._api_key_manager.get_api_key() as api_key:
            params["api_key"] = api_key
            params.update(query)

        try:
            with requests.get(
                    self.API_ENDPOINT_URL,
                    params=params
            ) as response:
                results = response.json()
                assert "profile" in results

        except (
            ConnectionError,
            json.decoder.JSONDecodeError,
            requests.exceptions.RequestException,
            urllib3.exceptions.HTTPError,
        ) as exception:
            # API hicups, let’s consider this batch
            # unsuccessful and start over
            raise ApiResponseError() from exception

        except AssertionError:
            # TODO: implement logging and report the response text + headers
            # if API hicups, return a stub data dict
            results = {
                "profile": {
                    "id": nsid
                }
            }

        return results["profile"]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Copyright (C) 2020 Christoph Fink, University of Helsinki
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 3
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, see <http://www.gnu.org/licenses/>.


"""Thread to complete missing data on user profiles."""


__all__ = ["UserProfileUpdaterThread"]


import threading
import time

import sqlalchemy

from .config import Config
from .databaseobjects import FlickrUser
from .exceptions import ApiResponseError
from .userprofiledownloader import UserProfileDownloader


class UserProfileUpdaterThread(threading.Thread):
    """Finds incomplete user profiles and downloads missing data from the flickr API."""

    MAX_RETRIES = 5  # once all users have been updated, retry this times (with 10 min breaks)

    def __init__(
            self,
            api_key_manager,
            partition=None
    ):
        """
        Intialize a UserProfileUpdateThread.

        Args:
            api_key_manager: instance of an ApiKeyManager
            partition (tuple of int): download the n-th of m parts of incomplete users

        """
        super().__init__()

        self.count = 0

        self._api_key_manager = api_key_manager
        try:
            part, number_of_partitions = partition
            assert part > 0
            assert part <= number_of_partitions
            self._bounds = (
                (part - 1) * 1.0 / number_of_partitions,
                part * 1.0 / number_of_partitions
            )
        except (
            AssertionError,
            TypeError
        ):
            self._bounds = None

        self.shutdown = threading.Event()

        with Config() as config:
            self._engine = sqlalchemy.create_engine(
                config["database_connection_string"]
            )

    @property
    def nsids_of_users_without_detailed_information(self):
        """Find nsid of incomplete user profiles."""
        # Find nsid of incomplete user profiles
        # We use join_date IS NULL, because after
        # updating a profile it will be "", so NULL is
        # a good way of finding “new” profiles
        with sqlalchemy.orm.Session(self._engine) as session:
            if self._bounds is None:
                nsids_of_users_without_detailed_information = (
                    session.query(FlickrUser.nsid).filter_by(join_date=None)
                )
            else:
                bounds = (
                    sqlalchemy.select(
                        sqlalchemy.sql.functions.percentile_disc(self._bounds[0])
                        .within_group(FlickrUser.id)
                        .label("lower"),
                        sqlalchemy.sql.functions.percentile_disc(self._bounds[1])
                        .within_group(FlickrUser.id)
                        .label("upper")
                    )
                    .select_from(FlickrUser)
                    .filter_by(join_date=None)
                    .cte()
                )
                nsids_of_users_without_detailed_information = (
                    session.query(FlickrUser.nsid)
                    .filter_by(join_date=None)
                    .where(FlickrUser.id.between(bounds.c.lower, bounds.c.upper))
                    .yield_per(1000)
                )

            for nsid, in nsids_of_users_without_detailed_information:
                yield nsid

    def run(self):
        """Get TimeSpans off todo_queue and download photos."""
        user_profile_downloader = UserProfileDownloader(self._api_key_manager)

        retries = 0

        while not (
                self.shutdown.is_set()
                or retries >= self.MAX_RETRIES
        ):
            for nsid in self.nsids_of_users_without_detailed_information:
                try:
                    with sqlalchemy.orm.Session(
                            self._engine
                    ) as session, session.begin():
                        flickr_user = (
                            FlickrUser.from_raw_api_data_flickrprofilegetprofile(
                                user_profile_downloader.get_profile_for_nsid(nsid)
                            )
                        )
                        session.merge(flickr_user)

                    self.count += 1

                except ApiResponseError:
                    # API returned some bogus/none-JSON data,
                    # let’s try again later
                    continue

                if self.shutdown.is_set():
                    break

            # once no incomplete user profiles remain,
            # wait for ten minutes before trying again;
            # wake up every 1/10 sec to check whether we
            # should shut down
            for _ in range(10 * 60 * 10):
                if self.shutdown.is_set():
                    break
                time.sleep(0.1)
            retries += 1



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Copyright (C) 2019 Christoph Fink, University of Helsinki
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 3
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, see <http://www.gnu.org/licenses/>.


"""A timer-released threading.Lock-like lock."""


__all__ = ["TimeoutLock"]


import threading
import time


class TimeoutLock:
    """A timer-released threading.Lock-like lock."""

    def __init__(self, timeout=1.0):
        """Intialize a timer-released threading.Lock-like lock."""
        self._lock = threading.Lock()
        self._lock_time = 0
        self.timeout = timeout

    def __enter__(self):
        """Acquire the lock using a contextmanager."""
        self.acquire()

    def __exit__(self, *_):
        """Release the lock using a contextmanager."""
        self.release()

    def acquire(self, blocking=True, timeout=-1):
        """Acquire the lock, blocking or non-blocking."""
        if self._lock.acquire(blocking=False):
            return True

        if blocking:
            wait_time = (self._lock_time + self.timeout) - time.time()
            if timeout == -1 or timeout > wait_time:
                if wait_time > 0:
                    time.sleep(wait_time)
                self._lock.release()
                return self.acquire()

        return False

    def release(self):
        """Release the lock."""
        self._lock_time = time.time()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Copyright (C) 2019 Christoph Fink, University of Helsinki
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 3
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, see <http://www.gnu.org/licenses/>.


"""Manages API keys (and their rate limit)."""


__all__ = ["ApiKeyManager"]


import contextlib
import queue


from .timeoutlock import TimeoutLock


class ApiKey(TimeoutLock):
    """An API key that manages its own rate limit."""

    def __init__(self, api_key, rate_limit_per_second):
        """
        Initialise an API key.

        Args:
            api_key: API key (mixed)
            rate_limit_per_second: How often this API key
                can be used per second
        """
        super().__init__(timeout=(1.0 / rate_limit_per_second))
        self._api_key = api_key

    def __enter__(self):
        """Return the API key, after timeout if over rate limit."""
        self.acquire()
        return self._api_key


class ApiKeyManager:
    """Manages API keys (and their rate limit)."""

    def __init__(self, api_keys=None, rate_limit_per_second=1.0):
        """Intialize an API key manager."""
        self._api_keys = queue.Queue()
        self.rate_limit_per_second = rate_limit_per_second

        self._has_api_keys = False
        if api_keys is not None:
            for api_key in api_keys:
                self.add_api_key(api_key)

    def add_api_key(self, api_key):
        """Add an API key to the manager."""
        self._api_keys.put(ApiKey(api_key, self.rate_limit_per_second))
        self._has_api_keys = True

    @contextlib.contextmanager
    def get_api_key(self):
        """Retrieve the next available API key."""
        if not self._has_api_keys:
            raise RuntimeError("No API keys configured")

        api_key = self._api_keys.get()
        with api_key as key:
            try:
                yield key
            finally:
                self._api_keys.put(api_key)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Copyright (C) 2019 Christoph Fink, University of Helsinki
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 3
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, see <http://www.gnu.org/licenses/>.

"""Import global configuration from a default location."""


__all__ = ["Config"]


import os
import os.path
import warnings

import yaml


class Config:
    """
    Global configuration object.

    Tries to load configuration from YAML files in default locations
    (/etc/{module}.yml, ~/.config/{module}.yml, %APPDATA%/{module}.yml,
    ${XDG_CONFIG_HOME}/{module}.yml).

    Configuration file location can be overriden by specifying
    `config_files` or a base name `config_filesBaseName` (replaces
    “{module}” in the file locations listed above)

    Arguments:
        content of the config file

    Args:
        config_files (list of str, optional): override configuration
            file locations
        config_files_basename (str): override configuration file names

    """

    # implementing Config() as a pseudo-singleton:
    # `_config` is a class attribute, it is shared
    # between instances of Config()

    _config = {}

    def __init__(
            self,
            config=None,
            config_files=None,
            config_files_basename=None
    ):
        """Initialise a Config object, load configuration from file."""
        if not self._config:
            self._config.update(
                self._load_config(
                    config_files,
                    config_files_basename
                )
            )
        if config is not None:
            self._config.update(config)

    def _load_config(
            self,
            config_files,
            config_files_basename
    ):
        config = {}

        if config_files is not None:
            if not isinstance(config_files, (list, tuple)):
                config_files = [config_files]

        elif config_files_basename is None:
            config_files_basename = self.__module__.split(".")[0]

            config_files = [
                "/etc/{:s}.yml".format(config_files_basename),
                os.path.abspath(
                    os.path.join(
                        (
                            os.environ.get('APPDATA')
                            or os.environ.get('XDG_CONFIG_HOME')
                            or os.path.join(os.environ['HOME'], '.config')
                        ),
                        "{:s}.yml".format(config_files_basename)
                    )
                )
            ]

        for config_file in config_files:
            try:
                config.update(
                    yaml.safe_load(
                        open(config_file, "r", encoding="utf-8")
                    )
                )
            except FileNotFoundError:
                pass

        if config == {}:
            warnings.warn(
                "No configuration found in files {}.".format(
                    ",".join(config_files)
                )
            )

        return config

    def __getitem__(self, pos):
        """Retrieve configuration value."""
        return self._config[pos]

    def __setitem__(self, pos, value):
        """Set configuration value."""
        self._config[pos] = value

    def __iter__(self):
        """Iterate over configuration file entries."""
        return iter(self._config)

    def __enter__(self):
        """Enter config context."""
        return self

    def __exit__(self, *_):
        """Exit config context."""
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Copyright (C) 2020 Christoph Fink, University of Helsinki
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 3
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, see <http://www.gnu.org/licenses/>.


"""Download all data covering a time span from the flickr API."""


__all__ = ["PhotoDownloader"]


import datetime
import json

import requests
import urllib3

from .exceptions import ApiResponseError, DownloadBatchIsTooLargeError


class PhotoDownloader:
    """Download all data covering a time span from the flickr API."""

    API_ENDPOINT_URL = "https://api.flickr.com/services/rest/"

    def __init__(self, timespan, api_key_manager):
        """Intialize an PhotoDownloader."""
        self._timespan = timespan
        self._api_key_manager = api_key_manager

    @property
    def photos(self):
        """Iterate over downloaded photos."""
        query = {
            "method": "flickr.photos.search",
            "format": "json",
            "nojsoncallback": 1,

            "per_page": 500,
            "has_geo": 1,
            "extras": ", ".join(
                ["description", "date_upload", "date_taken", "geo", "owner_name"]
            ),

            "min_upload_date": self._timespan.start.timestamp(),
            "max_upload_date": self._timespan.end.timestamp(),
            "sort": "date-posted-asc",
        }

        page = 1

        while True:
            query["page"] = page
            params = {}
            with self._api_key_manager.get_api_key() as api_key:
                params["api_key"] = api_key
                params.update(query)

                try:
                    with requests.get(
                            self.API_ENDPOINT_URL,
                            params=params
                    ) as response:
                        results = response.json()
                except (
                    ConnectionError,
                    json.decoder.JSONDecodeError,
                    requests.exceptions.RequestException,
                    urllib3.exceptions.HTTPError,
                ) as exception:
                    # API hicups, let’s consider this batch
                    # unsuccessful and start over
                    raise ApiResponseError() from exception

            try:
                num_photos = int(results["photos"]["total"])
            except TypeError:
                num_photos = 0

            if (
                    num_photos > 4000
                    and self._timespan.duration > datetime.timedelta(seconds=1)
            ):
                raise DownloadBatchIsTooLargeError(
                    (
                        "More than 4000 rows returned ({:d}), "
                        + "please specify a shorter time span."
                    ).format(num_photos)
                )

            for photo in results["photos"]["photo"]:

                # the flickr API is matching date_posted very fuzzily,
                # let’s not waste time with duplicates
                if (
                        datetime.datetime.fromtimestamp(
                            int(photo["dateupload"]),
                            tz=datetime.timezone.utc
                        ) > self._timespan.end
                ):
                    break

                yield photo

            page += 1
            if page > int(results["photos"]["pages"]):
                break



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Copyright (C) 2019 Christoph Fink, University of Helsinki
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 3
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, see <http://www.gnu.org/licenses/>.

"""Base classes to represent flickr posts and users."""


__all__ = [
    "FlickrPhoto",
    "FlickrUser"
]


import datetime

import geoalchemy2
import sqlalchemy
import sqlalchemy.ext.declarative
import sqlalchemy.ext.hybrid
import sqlalchemy.orm

from .config import Config


Base = sqlalchemy.ext.declarative.declarative_base()
config = Config()


class FlickrUser(Base):
    """ORM class to represent a flickr user."""

    __tablename__ = "users"

    id = sqlalchemy.Column(sqlalchemy.BigInteger)
    farm = sqlalchemy.Column(sqlalchemy.SmallInteger)
    nsid = sqlalchemy.Column(
        sqlalchemy.Text,
        sqlalchemy.Computed("id::TEXT || '@N0' || farm::TEXT")
    )

    name = sqlalchemy.Column(sqlalchemy.Text)
    first_name = sqlalchemy.Column(sqlalchemy.Text)
    last_name = sqlalchemy.Column(sqlalchemy.Text)
    real_name = sqlalchemy.Column(
        sqlalchemy.Text,
        sqlalchemy.Computed("first_name || ' ' || last_name")
    )

    city = sqlalchemy.Column(sqlalchemy.Text)
    country = sqlalchemy.Column(sqlalchemy.Text)
    hometown = sqlalchemy.Column(sqlalchemy.Text)

    occupation = sqlalchemy.Column(sqlalchemy.Text)
    description = sqlalchemy.Column(sqlalchemy.Text)

    join_date = sqlalchemy.Column(sqlalchemy.DateTime(timezone=True))

    website = sqlalchemy.Column(sqlalchemy.Text)
    facebook = sqlalchemy.Column(sqlalchemy.Text)
    twitter = sqlalchemy.Column(sqlalchemy.Text)
    tumblr = sqlalchemy.Column(sqlalchemy.Text)
    instagram = sqlalchemy.Column(sqlalchemy.Text)
    pinterest = sqlalchemy.Column(sqlalchemy.Text)

    photos = sqlalchemy.orm.relationship("FlickrPhoto", back_populates="user")

    __table_args__ = (
        sqlalchemy.PrimaryKeyConstraint("id", "farm"),
    )

    @classmethod
    def from_raw_api_data_flickrphotossearch(cls, data):
        """Initialise a new FlickrUser with a flickr.photos.search data dict."""
        user_id, farm = data["owner"].split("@N0")
        user_data = {
            "id": user_id,
            "farm": farm,
            "name": data["ownername"]
        }
        return cls(**user_data)

    @classmethod
    def from_raw_api_data_flickrprofilegetprofile(cls, data):
        """Initialise a new FlickrUser with a flickr.profile.getProfile data dict."""
        # the API does not always return all fields

        # "id" is the only field garantueed to be in the data
        # (because we add it ourselves in databaseobjects.py in case parsing fails)
        user_id, farm = data["id"].split("@N0")

        # "joindate" needs special attentation
        try:
            join_date = datetime.datetime.fromtimestamp(
                int(data["join_date"]),
                tz=datetime.timezone.utc
            )
        except KeyError:
            join_date = None

        user_data = {
            "id": user_id,
            "farm": farm,

            "join_date": join_date
        }

        # all the other fields can be added as they are (if they exist)
        for field in [
            "first_name",
            "last_name",

            "city",
            "country",
            "hometown",

            "occupation",
            "description",

            "website",
            "facebook",
            "twitter",
            "tumblr",
            "instagram",
            "pinterest"
        ]:
            try:
                user_data[field] = data[field]
            except KeyError:
                pass

        return cls(**user_data)

    def __str__(self):
        """Return a str representation."""
        return "<FlickrUser({:s}@N0{:s})>".format(self.id, self.farm)

    def __repr(self):
        """Return a str representation."""
        return str(self)


class FlickrPhoto(Base):
    """ORM class to represent a flickr photo (posts)."""

    __tablename__ = "photos"

    id = sqlalchemy.Column(sqlalchemy.BigInteger, primary_key=True)

    server = sqlalchemy.Column(sqlalchemy.Integer)
    secret = sqlalchemy.Column(sqlalchemy.LargeBinary)

    title = sqlalchemy.Column(sqlalchemy.Text)
    description = sqlalchemy.Column(sqlalchemy.Text)

    date_taken = sqlalchemy.Column(sqlalchemy.DateTime(timezone=True))
    date_posted = sqlalchemy.Column(sqlalchemy.DateTime(timezone=True))

    photo_url = sqlalchemy.Column(
        sqlalchemy.Text,
        sqlalchemy.Computed(
            "'https://live.staticflickr.com/' || server::TEXT || '/' || "
            + "id::TEXT || '_' || encode(secret, 'hex') || '_z.jpg'"
        )
    )
    page_url = sqlalchemy.Column(
        sqlalchemy.Text,
        sqlalchemy.Computed(
            "'https://www.flickr.com/photos/' || "
            + "user_id::TEXT || '@N0' || user_farm::TEXT || '/' || "
            + "id::TEXT || '/'"
        )
    )

    geom = sqlalchemy.Column(geoalchemy2.Geometry("POINT", 4326))

    user_id = sqlalchemy.Column(sqlalchemy.BigInteger, nullable=False)
    user_farm = sqlalchemy.Column(sqlalchemy.SmallInteger, nullable=False)

    user = sqlalchemy.orm.relationship("FlickrUser", back_populates="photos")

    __table_args__ = (
        sqlalchemy.ForeignKeyConstraint(
            ["user_id", "user_farm"],
            ["users.id", "users.farm"],
            "FlickrUser"
        ),
    )

    @classmethod
    def from_raw_api_data_flickrphotossearch(cls, data):
        """Initialise a new FlickrPhoto with a flickr.photos.search data dict."""
        # the API does not always return all fields
        # we need to figure out which ones we can use

        # and do quite a lot of clean-up because the flickr API
        # also returns fairly weird data, sometimes

        # another side effect is that we can initialise
        # with incomplete data (only id needed),
        # which helps with bad API responses

        photo_data = {}

        # "id" is the only field garantueed to be in the data
        # (because we add it ourselves in databaseobjects.py in case parsing fails)
        photo_data["id"] = data["id"]

        # server and secret are kinda straight-forward
        try:
            photo_data["server"] = data["server"]
        except KeyError:
            pass

        try:
            photo_data["secret"] = bytes.fromhex(data["secret"])
        except (
            ValueError,  # some non-hex character
            KeyError
        ):
            pass

        try:
            photo_data["title"] = data["title"]
        except KeyError:
            pass

        try:
            photo_data["description"] = data["description"]["_content"]
        except KeyError:
            pass

        # the dates need special attention
        try:
            photo_data["date_taken"] = datetime.datetime.fromisoformat(
                data["datetaken"]
            ).astimezone(datetime.timezone.utc)
        except ValueError:
            # there is weirdly quite a lot of photos with
            # date_taken "0000-01-01 00:00:00"
            # Year 0 does not exist, there’s 1BCE, then 1CE, nothing in between
            photo_data["date_taken"] = None
        except KeyError:
            # field does not exist in the dict we got
            pass

        try:
            photo_data["date_posted"] = datetime.datetime.fromtimestamp(
                int(data["dateupload"]),
                tz=datetime.timezone.utc
            )
        except KeyError:
            pass

        # geometry
        try:
            longitude = float(data["longitude"])
            latitude = float(data["latitude"])
            assert longitude != 0 and latitude != 0
            photo_data["geom"] = "SRID=4326;POINT({longitude:f} {latitude:f})".format(
                longitude=longitude,
                latitude=latitude
            )
        except (
            AssertionError,  # lon/lat is at exactly 0°N/S, 0°W/E -> bogus
            KeyError,        # not contained in API dict
            TypeError        # weird data returned
        ):
            pass

        # finally, the user
        # (let’s just delegate that to the FlickrUser constructor)
        photo_data["user"] = FlickrUser.from_raw_api_data_flickrphotossearch(data)

        return cls(**photo_data)

    def __str__(self):
        """Return a str representation."""
        return "<FlickrPhoto({:s})>".format(self.id)

    def __repr(self):
        """Return a str representation."""
        return str(self)


# Create tables in case we know where
if "database_connection_string" in config:
    engine = sqlalchemy.create_engine(config["database_connection_string"])
    Base.metadata.create_all(engine)
